/*
* JBoss, Home of Professional Open Source
* Copyright 2010, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.test.remoting.connection.identity;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.lang.reflect.Field;
import java.net.InetAddress;
import java.security.AccessController;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;
import java.util.ConcurrentModificationException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.management.MBeanServer;

import junit.framework.TestCase;

import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.jboss.logging.XLevel;
import org.jboss.remoting.Client;
import org.jboss.remoting.ConnectionListener;
import org.jboss.remoting.ConnectionValidator;
import org.jboss.remoting.InvocationRequest;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.Remoting;
import org.jboss.remoting.ServerInvocationHandler;
import org.jboss.remoting.ServerInvoker;
import org.jboss.remoting.callback.InvokerCallbackHandler;
import org.jboss.remoting.transport.Connector;
import org.jboss.remoting.transport.PortUtil;
import org.jboss.remoting.transport.socket.LRUPool;
import org.jboss.remoting.transport.socket.ServerSocketWrapper;
import org.jboss.remoting.transport.socket.ServerThread;
import org.jboss.remoting.transport.socket.SocketServerInvoker;
import org.jboss.remoting.util.SecurityUtility;


/**
 * Unit test for JBREM-1144.
 * 
 * @author <a href="mailto:ron.sigal@jboss.com">Ron Sigal</a>
 * @version $Rev$
 * <p>
 * Copyright Nov 17, 2010
 * </p>
 */
public class ServerIdentityTestCase extends TestCase
{
   private static Logger log = Logger.getLogger(ServerIdentityTestCase.class);
   
   protected String host;
   protected int port;
   protected String locatorURI;
   protected InvokerLocator serverLocator;
   protected Connector connector;
   protected TestInvocationHandler invocationHandler;
   protected ByteArrayOutputStream baosOut;
   protected PrintStream originalOutPrintStream;
   protected ByteArrayOutputStream baosErr;
   protected PrintStream originalErrPrintStream;
   
   public void setUp() throws Exception
   {
      originalOutPrintStream = System.out;
      baosOut = new ByteArrayOutputStream();
      PrintStream ps = new PrintStream(baosOut);
      setOut(ps);
      
      originalErrPrintStream = System.err;
      baosErr = new ByteArrayOutputStream();
      ps = new PrintStream(baosErr);
      setErr(ps);
      
      Logger.getLogger("org.jboss.remoting").setLevel(XLevel.TRACE);
      Logger.getLogger("org.jboss.test.remoting").setLevel(Level.INFO);
      String pattern = "[%d{ABSOLUTE}] [%t] %5p (%F:%L) - %m%n";
      PatternLayout layout = new PatternLayout(pattern);
      ConsoleAppender consoleAppender = new ConsoleAppender(layout);
      Logger.getRootLogger().addAppender(consoleAppender);
      
      host = InetAddress.getLocalHost().getHostAddress();
      port = PortUtil.findFreePort(host);
   }

   
   public void tearDown()
   {
   }
   
   
   public void testServerIdentityWithoutLeasing() throws Throwable
   {
      log.info("entering " + getName());
      
      // Start server.
      setupServer(false);
      
      // Create client. Adding "dummy=dummy" assures that InvokerRegistry will not create a
      // LocalClientInvoker for the ConnectionValidator.
      InvokerLocator clientLocator = new InvokerLocator(locatorURI + "&dummy=dummy");
      HashMap clientConfig = new HashMap();
      clientConfig.put(InvokerLocator.FORCE_REMOTE, "true");
      clientConfig.put("numberOfCallRetries", "3"); // for ConnectionValidator
      addExtraClientConfig(clientConfig);
      Client client = new Client(clientLocator, clientConfig);
      client.connect();
      log.info("client is connected");
      
      // Test connection.
      assertEquals("abc", client.invoke("abc"));
      log.info("connection is good");
      
      // Install connection listener.
      TestConnectionListener listener = new TestConnectionListener();
      HashMap metadata = new HashMap();
      metadata.put(Remoting.USE_SERVER_CONNECTION_IDENTITY, "true");
      metadata.put(ConnectionValidator.VALIDATOR_PING_PERIOD, "10000");
      metadata.put(ConnectionValidator.VALIDATOR_PING_TIMEOUT, "10000");
      client.addConnectionListener(listener, metadata);
      log.info(this + " added connection listener: " + listener);
      // Allow time to get serverId of first server.
      Thread.sleep(15000);
      
      // Verify listener is notified if server bounces.
      shutdownServer();
      log.info("SHUT DOWN SERVER");
      setupServer(false);
      log.info("SET UP NEW SERVER");
      Thread.sleep(10000);
      log.info(this + " listener.connectionFailed: " + listener.connectionFailed);
      assertTrue(listener.connectionFailed);
      
      setOut(originalOutPrintStream);
      String sOut = new String(baosOut.toByteArray());
      System.out.println(sOut);
      setErr(originalErrPrintStream);
      String sErr = new String(baosErr.toByteArray());
      System.out.println(sErr);
      assertTrue(sOut.indexOf("detected new serverId:") > -1 || sErr.indexOf("detected new serverId:") > -1);
     
      client.disconnect();
      shutdownServer();
     
      log.info(getName() + " PASSES");
   }
   
   public void testServerIdentityWithLeasingNewServer() throws Throwable
   {
      log.info("entering " + getName());
      
      // Start server.
      setupServer(true);
      
      // Create client. Adding "dummy=dummy" assures that InvokerRegistry will not create a
      // LocalClientInvoker for the ConnectionValidator.
      InvokerLocator clientLocator = new InvokerLocator(locatorURI + "&dummy=dummy");
      HashMap clientConfig = new HashMap();
      clientConfig.put(InvokerLocator.FORCE_REMOTE, "true");
      clientConfig.put("numberOfCallRetries", "3"); // for ConnectionValidator
      addExtraClientConfig(clientConfig);
      Client client = new Client(clientLocator, clientConfig);
      client.connect();
      log.info("client is connected");
      
      // Test connection.
      assertEquals("abc", client.invoke("abc"));
      log.info("connection is good");
      
      // Install connection listener.
      TestConnectionListener listener = new TestConnectionListener();
      HashMap metadata = new HashMap();
      metadata.put(Remoting.USE_SERVER_CONNECTION_IDENTITY, "true");
      metadata.put(ConnectionValidator.VALIDATOR_PING_PERIOD, "10000");
      metadata.put(ConnectionValidator.VALIDATOR_PING_TIMEOUT, "10000");
      client.addConnectionListener(listener, metadata);
      log.info(this + " added connection listener: " + listener);
      // Allow time to get serverId of first server.
      Thread.sleep(15000);
      
      // Verify listener is notified if server bounces.
      shutdownServer();
      log.info("SHUT DOWN SERVER");
      setupServer(true);
      log.info("SET UP NEW SERVER");
      Thread.sleep(15000);
      log.info(this + " listener.connectionFailed: " + listener.connectionFailed);
      assertTrue(listener.connectionFailed);
      
      setOut(originalOutPrintStream);
      String sOut = new String(baosOut.toByteArray());
      System.out.println(sOut);
      setErr(originalErrPrintStream);
      String sErr = new String(baosErr.toByteArray());
      System.out.println(sErr);
      assertTrue(sOut.indexOf("detected new serverId:") > -1 || sErr.indexOf("detected new serverId:") > -1);
     
      client.disconnect();
      shutdownServer();
     
      log.info(getName() + " PASSES");
   }
   
   
   public void testServerIdentityWithLeasingNewLease() throws Throwable
   {
      log.info("entering " + getName());
      
      // Start server.
      setupServer(true);
      
      // Create client. Adding "dummy=dummy" assures that InvokerRegistry will not create a
      // LocalClientInvoker for the ConnectionValidator.
      InvokerLocator clientLocator = new InvokerLocator(locatorURI + "&dummy=dummy");
      HashMap clientConfig = new HashMap();
      clientConfig.put(InvokerLocator.FORCE_REMOTE, "true");
      clientConfig.put("numberOfCallRetries", "3"); // for ConnectionValidator
      addExtraClientConfig(clientConfig);
      Client client = new Client(clientLocator, clientConfig);
      client.connect();
      log.info("client is connected");
      
      // Test connection.
      assertEquals("abc", client.invoke("abc"));
      log.info("connection is good");
      
      // Install connection listener.
      TestConnectionListener listener = new TestConnectionListener();
      HashMap metadata = new HashMap();
      metadata.put(Remoting.USE_SERVER_CONNECTION_IDENTITY, "true");
      metadata.put(ConnectionValidator.VALIDATOR_PING_PERIOD, "10000");
      metadata.put(ConnectionValidator.VALIDATOR_PING_TIMEOUT, "10000");
      client.addConnectionListener(listener, metadata);
      log.info(this + " added connection listener: " + listener);
      // Allow time to get serverId of first server.
      Thread.sleep(15000);
      
      // Verify listener is notified if Lease is destroyed.
      Field field = ServerInvoker.class.getDeclaredField("clientLeases");
      field.setAccessible(true);
      Map clientLeases = (Map) field.get(connector.getServerInvoker());
      clientLeases.clear();
      log.info("REMOVED LEASE");
      Thread.sleep(15000);
      log.info(this + " listener.connectionFailed: " + listener.connectionFailed);
      assertTrue(listener.connectionFailed);
      
      setOut(originalOutPrintStream);
      String sOut = new String(baosOut.toByteArray());
      setErr(originalErrPrintStream);
      String sErr = new String(baosErr.toByteArray());
      if (!(sOut.indexOf("detected new serverId:") > -1 || sErr.indexOf("detected new serverId:") > -1))
      {
         System.out.println(sOut);
         System.out.println(sErr);
      }
      assertTrue(sOut.indexOf("detected new serverId:") > -1 || sErr.indexOf("detected new serverId:") > -1);
     
      client.disconnect();
      shutdownServer();
     
      log.info(getName() + " PASSES");
   }
   
   
   protected String getTransport()
   {
      return "socket";
   }
   
   
   protected void addExtraClientConfig(Map config) {}
   protected void addExtraServerConfig(Map config) {}
   

   private void setOut(final PrintStream ps)
   {
      if (SecurityUtility.skipAccessControl())
      {
         System.setOut(ps);
         return;
      }
      
      try
      {
         AccessController.doPrivileged( new PrivilegedExceptionAction()
         {
            public Object run() throws Exception
            {
               System.setOut(ps);
               return null;
            }
         });
      }
      catch (PrivilegedActionException e)
      {
         throw (RuntimeException) e.getCause();
      }
   }
   
   
   private void setErr(final PrintStream ps)
   {
      if (SecurityUtility.skipAccessControl())
      {
         System.setErr(ps);
         return;
      }
      
      try
      {
         AccessController.doPrivileged( new PrivilegedExceptionAction()
         {
            public Object run() throws Exception
            {
               System.setErr(ps);
               return null;
            }
         });
      }
      catch (PrivilegedActionException e)
      {
         throw (RuntimeException) e.getCause();
      }
   }
   
   
   protected void setupServer(boolean useLeasing) throws Exception
   {
      locatorURI = getTransport() + "://" + host + ":" + port + "?" + Remoting.USE_SERVER_CONNECTION_IDENTITY + "=true";
      locatorURI += "&" + SocketServerInvoker.CHECK_CONNECTION_KEY + "=true";
      if (useLeasing)
      {
         locatorURI += "&" + InvokerLocator.CLIENT_LEASE + "=true";
         locatorURI += "&" + InvokerLocator.CLIENT_LEASE_PERIOD + "=2000";
      }
      String metadata = System.getProperty("remoting.metadata");
      if (metadata != null)
      {
         locatorURI += "&" + metadata;
      }
      serverLocator = new InvokerLocator(locatorURI);
      log.info("Starting remoting server with locator uri of: " + locatorURI);
      HashMap config = new HashMap();
      config.put(InvokerLocator.FORCE_REMOTE, "true");
      addExtraServerConfig(config);
      connector = new Connector(serverLocator, config);
      connector.create();
      invocationHandler = new TestInvocationHandler();
      connector.addInvocationHandler("test", invocationHandler);
      if (useLeasing)
      {
         connector.addConnectionListener(new TestConnectionListener());
      }
      connector.start();
   }
   
   
   protected void shutdownServer() throws Exception
   {
      if (connector != null)
      {
         // Remoting versions 1.x and 2.2.x don't necessarily shut down all of their
         // ServerThreads, so an exiting connection could connect to a ServerThread
         // associated with the old ServerInvoker.
         SocketServerInvoker invoker = (SocketServerInvoker) connector.getServerInvoker();
         Field clientpoolField = SocketServerInvoker.class.getDeclaredField("clientpool");
         clientpoolField.setAccessible(true);
         Field socketWrapperField = ServerThread.class.getDeclaredField("socketWrapper");
         socketWrapperField.setAccessible(true);
         LRUPool clientpool = (LRUPool) clientpoolField.get(invoker);
         Set threads = clientpool.getContents();
         Iterator it = threads.iterator();
         while (it.hasNext())
         {
            try
            {
               ServerThread t = (ServerThread) it.next();
               ServerSocketWrapper socketWrapper = (ServerSocketWrapper) socketWrapperField.get(t);
               socketWrapper.close();
            }
            catch (ConcurrentModificationException e)
            {
               log.info(this + " caught " + e.getMessage());
            }
         }
         connector.stop();
      }
   }
   
   
   static class TestInvocationHandler implements ServerInvocationHandler
   {
      public void addListener(InvokerCallbackHandler callbackHandler) {}
      public Object invoke(final InvocationRequest invocation) throws Throwable
      {
         return invocation.getParameter();
      }
      public void removeListener(InvokerCallbackHandler callbackHandler) {}
      public void setMBeanServer(MBeanServer server) {}
      public void setInvoker(ServerInvoker invoker) {}
   }

   static class TestConnectionListener implements ConnectionListener
   {
      public boolean connectionFailed;
      public Throwable throwable;
      
      public void handleConnectionException(Throwable throwable, Client client)
      {
         connectionFailed = true;
         this.throwable = throwable;
         log.info(this + " received connection notification: connectionFailed: " + connectionFailed);
      }
      
   }
}