/*
* JBoss, Home of Professional Open Source
* Copyright 2009, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.test.remoting.stream.local;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.InetAddress;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import javax.management.MBeanServer;

import junit.framework.TestCase;

import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.jboss.remoting.Client;
import org.jboss.remoting.InvocationRequest;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.InvokerRegistry;
import org.jboss.remoting.ServerInvoker;
import org.jboss.remoting.callback.InvokerCallbackHandler;
import org.jboss.remoting.stream.StreamInvocationHandler;
import org.jboss.remoting.transport.Connector;
import org.jboss.remoting.transport.PortUtil;
import org.jboss.remoting.transport.local.LocalServerInvoker;
import org.jboss.remoting.transport.socket.SocketServerInvoker;


/**
 * Unit test for JBREM-1288.
 * 
 * @author <a href="mailto:ron.sigal@jboss.com">Ron Sigal</a>
 * @version $Rev$
 * <p>
 * Copyright Apr 29, 2011
 * </p>
 */
public class LocalStreamTestCase extends TestCase
{
   private static Logger log = Logger.getLogger(LocalStreamTestCase.class);
   
   private static boolean firstTime = true;
   
   protected String host;
   protected int port;
   protected String locatorURI;
   protected InvokerLocator serverLocator;
   protected Connector connector;
   protected File testFile = null;
   protected FileInputStream fileInput = null;
   protected TestStreamInvocationHandler invocationHandler;

   
   public void setUp() throws Exception
   {
      if (firstTime)
      {
         firstTime = false;
         Logger.getLogger("org.jboss.remoting").setLevel(Level.INFO);
         Logger.getLogger("org.jboss.test.remoting").setLevel(Level.INFO);
         String pattern = "[%d{ABSOLUTE}] [%t] %5p (%F:%L) - %m%n";
         PatternLayout layout = new PatternLayout(pattern);
         ConsoleAppender consoleAppender = new ConsoleAppender(layout);
         Logger.getRootLogger().addAppender(consoleAppender);  
      }
   }

   
   public void tearDown()
   {
   }
   
   
   public void testLocalStream() throws Throwable
   {
      log.info("entering " + getName());
      
      // Start server.
      setupServer();
      
      // Create client.
      InvokerLocator clientLocator = new InvokerLocator(locatorURI);
      HashMap clientConfig = new HashMap();
      addExtraClientConfig(clientConfig);
      Client client = new Client(clientLocator, "test_stream", clientConfig);
      client.connect();
      log.info("client is connected");
      
      // Test sending stream.
      URL fileURL = this.getClass().getResource("../test.txt");
      if(fileURL == null)
      {
         throw new Exception("Can not find file test.txt");
      }
      testFile = new File(fileURL.getFile());
      fileInput = new FileInputStream(testFile);

      String param = "foobar";
      long fileLength = testFile.length();
      log.info("File size = " + fileLength);
      Object ret = client.invoke(fileInput, param);

      Map responseMap = (Map)ret;
      String subSys = (String)responseMap.get("subsystem");
      String clientId = (String)responseMap.get("clientid");
      String paramVal = (String)responseMap.get("paramval");

      assertEquals("test_stream".toUpperCase(), subSys);
      assertEquals(client.getSessionId(), clientId);
      assertEquals("foobar", paramVal);

      Object response = client.invoke("get_size");
      int returnedFileLength = ((Integer) response).intValue();
      log.info("Invocation response: " + response);
      log.info("returnedFileLength: " + returnedFileLength);
      assertEquals(fileLength, returnedFileLength);
      
      // Test for LocalServerInvoker.
      ServerInvoker[] serverInvokers = InvokerRegistry.getServerInvokers();
      assertEquals(2, serverInvokers.length);
      assertTrue(serverInvokers[0] instanceof SocketServerInvoker || serverInvokers[1] instanceof SocketServerInvoker);
      assertTrue(serverInvokers[0] instanceof LocalServerInvoker || serverInvokers[1] instanceof LocalServerInvoker);
      
      client.disconnect();
      shutdownServer();
      log.info(getName() + " PASSES");
   }
   
   
   protected String getTransport()
   {
      return "socket";
   }

      
   protected void addExtraClientConfig(Map config) {}
   protected void addExtraServerConfig(Map config) {}
   

   protected void setupServer() throws Exception
   {
      host = InetAddress.getLocalHost().getHostAddress();
      port = PortUtil.findFreePort(host);
      locatorURI = getTransport() + "://" + host + ":" + port;
      String metadata = System.getProperty("remoting.metadata");
      if (metadata != null)
      {
         locatorURI += "/?" + metadata;
      }
      serverLocator = new InvokerLocator(locatorURI);
      log.info("Starting remoting server with locator uri of: " + locatorURI);
      HashMap config = new HashMap();
      config.put(InvokerLocator.FORCE_REMOTE, "true");
      addExtraServerConfig(config);
      connector = new Connector(serverLocator, config);
      connector.create();
      invocationHandler = new TestStreamInvocationHandler();
      connector.addInvocationHandler("test_stream", invocationHandler);
      connector.start();
   }
   
   
   protected void shutdownServer() throws Exception
   {
      if (connector != null)
         connector.stop();
   }
   
   
   public static class TestStreamInvocationHandler implements StreamInvocationHandler
   {
      private InputStream stream = null;

      private int streamSize = 0;

      public Object invoke(InvocationRequest invocation) throws Throwable
      {
         // Print out the invocation request
         log.info("Invocation request is: " + invocation.getParameter());

         // Just going to return static string as this is just simple example code.
         return new Integer(streamSize);
      }

      public void addListener(InvokerCallbackHandler callbackHandler)
      {
      }
      public void removeListener(InvokerCallbackHandler callbackHandler)
      {
      }
      public void setMBeanServer(MBeanServer server)
      {
      }
      public void setInvoker(ServerInvoker invoker)
      {
      }

      public Object handleStream(InputStream stream, InvocationRequest param)
      {
         this.stream = stream;

         try
         {
            ByteArrayOutputStream out = new ByteArrayOutputStream();
            byte buf[] = new byte[4096];
            while(true)
            {
               int c = this.stream.read(buf);
               if(c < 0)
               {
                  break;
               }
               out.write(buf, 0, c);
            }
            byte[] bytes = out.toByteArray();
            streamSize = bytes.length;
            log.info("Read stream.  Length is: " + streamSize);
         }
         catch(IOException e)
         {
            e.printStackTrace();
         }
         // Not closing the InputStream to keep the StreamServer alive.
//         finally
//         {
//            try
//            {
////               log.info("closing stream");
////               stream.close();
//            }
//            catch(IOException e)
//            {
//               log.error("error closing stream", e);
//            }
//         }
         // build return map
         Map retMap = new HashMap();
         retMap.put("subsystem", param.getSubsystem());
         retMap.put("clientid", param.getSessionId());
         retMap.put("paramval", param.getParameter());
         return retMap;
      }
   }
}