/*
* JBoss, Home of Professional Open Source
* Copyright 2009, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.test.remoting.tests;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Field;
import java.net.InetAddress;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;

import javax.management.MBeanServer;

import junit.framework.TestCase;

import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.jboss.remoting.Client;
import org.jboss.remoting.InvocationRequest;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.ServerInvoker;
import org.jboss.remoting.callback.InvokerCallbackHandler;
import org.jboss.remoting.stream.StreamHandler;
import org.jboss.remoting.stream.StreamInvocationHandler;
import org.jboss.remoting.transport.Connector;
import org.jboss.remoting.transport.PortUtil;
import org.jboss.remoting.transport.socket.MicroSocketClientInvoker;


public class StreamingTest extends TestCase
{
   private static Logger log = Logger.getLogger(StreamingTest.class);
   
   private static boolean firstTime = true;
   
   protected String host;
   protected int port;
   protected String locatorURI;
   protected InvokerLocator serverLocator;
   protected Connector connector;
   protected TestStreamInvocationHandler invocationHandler;
   protected Connector streamConnector;
   
   private String localFileName = "sample.txt";
   private String remoteFileName = "server_sample.txt";

   
   public void setUp() throws Exception
   {
      if (firstTime)
      {
         firstTime = false;
         Logger.getLogger("org.jboss.remoting").setLevel(Level.INFO);
         Logger.getLogger("org.jboss.test.remoting").setLevel(Level.INFO);
         String pattern = "[%d{ABSOLUTE}] [%t] %5p (%F:%L) - %m%n";
         PatternLayout layout = new PatternLayout(pattern);
         ConsoleAppender consoleAppender = new ConsoleAppender(layout);
         Logger.getRootLogger().addAppender(consoleAppender);  
      }
   }

   
   public void tearDown()
   {
   }
   
   
   public void testMethod() throws Throwable
   {
      log.info("entering " + getName());
      
      // Start server.
      setupServer();
      
      // Create client.
      InvokerLocator clientLocator = new InvokerLocator(locatorURI);
      HashMap clientConfig = new HashMap();
      clientConfig.put(InvokerLocator.FORCE_REMOTE, "true");
      addExtraClientConfig(clientConfig);
      Client client = new Client(clientLocator, clientConfig);
      client.connect();
      log.info("client is connected");

      URL fileURL = this.getClass().getResource(localFileName);
      System.out.println("looking for file at " + fileURL);
      if(fileURL == null)
      {
         throw new Exception("Can not find file " + localFileName);
      }
      File testFile = new File(fileURL.getFile());
      FileInputStream fileInput = new FileInputStream(testFile);

      System.out.println("Sending input stream for file " + localFileName + " to server.");
      setupStreamServer();
      Object ret = client.invoke(fileInput, remoteFileName, streamConnector);

      long fileLength = testFile.length();
      System.out.println("Size of file sample.txt is " + fileLength);
      System.out.println("Server returned " + ret + " as the size of the file read.");

      Field field = TestStreamInvocationHandler.class.getDeclaredField("streamHandler");
      field.setAccessible(true);
      StreamHandler streamHandler = (StreamHandler) field.get(this.invocationHandler);
      field = StreamHandler.class.getDeclaredField("streamClient");
      field.setAccessible(true);
      Client streamClient = (Client) field.get(streamHandler);
      log.info("streamClient is disconnected: " + streamClient.isConnected());
      log.info("streamClient invoker: " + streamClient.getInvoker());
      assertFalse(streamClient.isConnected());
      assertNull(streamClient.getInvoker());
      
      // Give StreamServer time to close.
      Thread.sleep(4000);
      
      fileInput.close();
      client.disconnect();
      shutdownStreamServer();
      shutdownServer();
      log.info(getName() + " PASSES");
   }
   
   
   protected String getTransport()
   {
      return "socket";
   }
   
   
   protected void addExtraClientConfig(Map config) {}
   protected void addExtraServerConfig(Map config) {}
   

   protected void setupServer() throws Exception
   {
      host = InetAddress.getLocalHost().getHostAddress();
      port = PortUtil.findFreePort(host);
      locatorURI = getTransport() + "://" + host + ":" + port;
      String metadata = System.getProperty("remoting.metadata");
      if (metadata != null)
      {
         locatorURI += "/?" + metadata;
      }
      serverLocator = new InvokerLocator(locatorURI);
      log.info("Starting remoting server with locator uri of: " + locatorURI);
      HashMap config = new HashMap();
      config.put(InvokerLocator.FORCE_REMOTE, "true");
      addExtraServerConfig(config);
      connector = new Connector(serverLocator, config);
      connector.create();
      invocationHandler = new TestStreamInvocationHandler();
      connector.addInvocationHandler("test_stream", invocationHandler);
      connector.start(true);
   }
   
   
   protected void shutdownServer() throws Exception
   {
      if (connector != null)
         connector.stop();
   }
   
   
   void setupStreamServer() throws Exception
   {
      String host = InetAddress.getLocalHost().getHostAddress();
      int port = PortUtil.findFreePort(host);
      String locatorURI = getTransport() + "://" + host + ":" + port + "/?" + InvokerLocator.FORCE_REMOTE + "=true";
      InvokerLocator locator = new InvokerLocator(locatorURI);
      System.out.println("StreamServer locator: " + locator);
      streamConnector = new Connector(locator);
      streamConnector.start();
   }
   
   
   
   protected void shutdownStreamServer() throws Exception
   {
      if (streamConnector != null)
         streamConnector.stop();
   }
   
   
   public static class TestStreamInvocationHandler implements StreamInvocationHandler
   {
      StreamHandler streamHandler;
      private long streamSize = 0;

      /**
       * takes the incoming stream and writes out to a file specified by the other param specified.
       * will return the size of the file.
       *
       * @param stream
       * @param param
       * @return
       */
      public Object handleStream(InputStream stream, InvocationRequest param)
      {
         streamHandler = (StreamHandler) stream;
         
         try
         {
            String fileName = (String)param.getParameter();
            System.out.println("Received input stream from client to write out to file " + fileName);
            File newFile = new File(fileName);
            if(!newFile.exists())
            {
               newFile.createNewFile();
            }

            FileOutputStream out = new FileOutputStream(newFile, false);

            byte buf[] = new byte[4096];
            while(true)
            {
               int c = stream.read(buf);
               if(c < 0)
               {
                  break;
               }
               out.write(buf, 0, c);
            }

            out.flush();
            out.close();

            streamSize = newFile.length();
            System.out.println("New file " + fileName + " has been written out to " + newFile.getAbsolutePath());
            System.out.println("Size of " + newFile.getAbsolutePath() + " is " + streamSize);
         }
         catch(Throwable e)
         {
            e.printStackTrace();
         }
         finally
         {
            try
            {
               stream.close();
            }
            catch(IOException e)
            {
               log.info("unable to close " + stream);
            }
         }
         return new Long(streamSize);
      }
      
      public Object invoke(InvocationRequest invocation) throws Throwable
      {
         // Print out the invocation request
         System.out.println("Invocation request is: " + invocation.getParameter());

         // Return the size of the file already streamed to the server (and written to disk).
         return new Long(streamSize);
      }
      
      public void addListener(InvokerCallbackHandler callbackHandler)
      {
      }
      public void removeListener(InvokerCallbackHandler callbackHandler)
      {
      }
      public void setMBeanServer(MBeanServer server)
      {
      }
      public void setInvoker(ServerInvoker invoker)
      {
      }
   }
}