/*
 * JBoss, Home of Professional Open Source.
 * Copyright 2009, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags. See the copyright.txt file in the
 * distribution for a full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

package org.jboss.test.remoting.transport.coyote;

import java.lang.reflect.Field;
import java.net.InetAddress;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.management.MBeanServer;

import junit.framework.TestCase;

import org.apache.catalina.core.StandardThreadExecutor;
import org.apache.coyote.http11.Http11Protocol;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.jboss.remoting.InvocationRequest;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.ServerInvocationHandler;
import org.jboss.remoting.ServerInvoker;
import org.jboss.remoting.callback.InvokerCallbackHandler;
import org.jboss.remoting.transport.Connector;
import org.jboss.remoting.transport.PortUtil;
import org.jboss.remoting.transport.coyote.CoyoteInvoker;


/**
 * Unit test for JBREM-1283
 * 
 * @author <a href="ron.sigal@jboss.com">Ron Sigal</a>
 * @version $Revision: 1.1 $
 * <p>
 * Copyright Apr 13, 2011
 */
public class SetAttributeTestCase extends TestCase
{
   private static Logger log = Logger.getLogger(SetAttributeTestCase.class);
   
   private static boolean firstTime = true;
   
   protected String host;
   protected int port;
   protected String locatorURI;
   protected InvokerLocator serverLocator;
   protected Connector connector;
   protected TestInvocationHandler invocationHandler;

   
   public void setUp() throws Exception
   {
      if (firstTime)
      {
         firstTime = false;
         Logger.getLogger("org.jboss.remoting").setLevel(Level.INFO);
         Logger.getLogger("org.jboss.test.remoting").setLevel(Level.INFO);
         String pattern = "[%d{ABSOLUTE}] [%t] %5p (%F:%L) - %m%n";
         PatternLayout layout = new PatternLayout(pattern);
         ConsoleAppender consoleAppender = new ConsoleAppender(layout);
         Logger.getRootLogger().addAppender(consoleAppender);  
      }
   }

   
   public void tearDown()
   {
   }
   
   
   public void testSetProperties() throws Throwable
   {
      log.info("entering " + getName());
      
      // Start server.
      host = InetAddress.getLocalHost().getHostAddress();
      port = PortUtil.findFreePort(host);
      locatorURI = getTransport() + "://" + host + ":" + port;
      String metadata = System.getProperty("remoting.metadata");
      if (metadata != null)
      {
         locatorURI += "/?" + metadata;
      }
      serverLocator = new InvokerLocator(locatorURI);
      log.info("Starting remoting server with locator uri of: " + locatorURI);
      HashMap config = new HashMap();
      config.put(InvokerLocator.FORCE_REMOTE, "true");
        
      // Attributes set directly on Http11Protocol
      config.put("maxSavePostSize", "3");
      config.put("keepAlive", "false");
      config.put("compression", "on");
      config.put("testObject", new TestObject(5));
      
      // Attributes set on Http11Protocol.endpoint
      config.put("soTimeout", "7");
      config.put("tcpNoDelay", "true");
      config.put("address", InetAddress.getLocalHost());      

      // Executor
      String s = "daemon=false,maxIdleTime=11, minSpareThreads =13,maxThreads = 17 ,";
      s += " name= test ,namePrefix=prefix,threadPriority=" + (Thread.NORM_PRIORITY + 1);
      s += ", bogus=fake";
      config.put("executor", s);
      
      connector = new Connector(serverLocator, config);
      connector.create();
      invocationHandler = new TestInvocationHandler();
      connector.addInvocationHandler("test", invocationHandler);
      connector.start();
      CoyoteInvoker invoker = (CoyoteInvoker) connector.getServerInvoker();
      Field field = CoyoteInvoker.class.getDeclaredField("protocolHandlers");
      field.setAccessible(true);
      List protocolHandlers = (List) field.get(invoker);
      Http11Protocol protocolHandler = (Http11Protocol) protocolHandlers.iterator().next();
      log.info("maxSavePostSize:   " + protocolHandler.getMaxSavePostSize());
      log.info("keepAlive:         " + protocolHandler.getKeepAlive());
      log.info("compression:       " + protocolHandler.getCompression());
      log.info("testObject:        " + protocolHandler.getAttribute("testObject"));
      
      log.info("soTimeout:         " + protocolHandler.getSoTimeout()); 
      log.info("tcpNoDelay:        " + protocolHandler.getTcpNoDelay());
      log.info("address:           " + protocolHandler.getAddress());
      
      assertEquals(new Integer(3), new Integer(protocolHandler.getMaxSavePostSize()));
      assertEquals(Boolean.FALSE, new Boolean(protocolHandler.getKeepAlive()));
      assertEquals("on", protocolHandler.getCompression());
      assertEquals(new TestObject(5), protocolHandler.getAttribute("testObject"));
      
      assertEquals(new Integer(7), new Integer(protocolHandler.getSoTimeout()));
      assertEquals(Boolean.TRUE, new Boolean(protocolHandler.getTcpNoDelay()));
      assertEquals(InetAddress.getLocalHost(), protocolHandler.getAddress());
      
      StandardThreadExecutor executor = (StandardThreadExecutor) protocolHandler.getExecutor();
      log.info("daemon:          " + executor.isDaemon());
      log.info("maxIdleTime:     " + executor.getMaxIdleTime());
      log.info("minSpareThreads: " + executor.getMinSpareThreads());
      log.info("maxThreads:      " + executor.getMaxThreads());
      log.info("name:            " + executor.getName());
      log.info("namePrefix:      " + executor.getNamePrefix());
      log.info("threadPriority:  " + executor.getThreadPriority());
      
      assertFalse(executor.isDaemon());
      assertEquals(11, executor.getMaxIdleTime());
      assertEquals(13, executor.getMinSpareThreads());
      assertEquals(17, executor.getMaxThreads());
      assertEquals("test", executor.getName());
      assertEquals("prefix", executor.getNamePrefix());
      assertEquals(Thread.NORM_PRIORITY + 1, executor.getThreadPriority());
      
      if (connector != null)
         connector.stop();
      log.info(getName() + " PASSES");
   }
   
   
   protected String getTransport()
   {
      return "http";
   }
   
   
   protected void addExtraClientConfig(Map config) {}
   protected void addExtraServerConfig(Map config) {}
   
   static class TestInvocationHandler implements ServerInvocationHandler
   {
      public void addListener(InvokerCallbackHandler callbackHandler) {}
      public Object invoke(final InvocationRequest invocation) throws Throwable
      {
         return invocation.getParameter();
      }
      public void removeListener(InvokerCallbackHandler callbackHandler) {}
      public void setMBeanServer(MBeanServer server) {}
      public void setInvoker(ServerInvoker invoker) {}
   }
   
   static class TestObject
   {
      private int secret;
      public TestObject(int secret) {this.secret = secret;}
      public boolean equals(Object o)
      {
         if (!(o instanceof TestObject)) return false;
         return this.secret == ((TestObject) o).secret;
      }
      public String toString()
      {
         return "TestObject(" + secret + ")";
      }
   }
}