/*
 * JBoss, Home of Professional Open Source
 * Copyright 2010, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jboss.test.remoting.versioning.identity;

import java.lang.reflect.Field;
import java.net.InetAddress;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.management.MBeanServer;

import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.jboss.jrunit.extensions.ServerTestCase;
import org.jboss.logging.XLevel;
import org.jboss.remoting.Client;
import org.jboss.remoting.ConnectionListener;
import org.jboss.remoting.InvocationRequest;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.Remoting;
import org.jboss.remoting.ServerInvocationHandler;
import org.jboss.remoting.ServerInvoker;
import org.jboss.remoting.callback.InvokerCallbackHandler;
import org.jboss.remoting.transport.Connector;
import org.jboss.remoting.transport.socket.LRUPool;
import org.jboss.remoting.transport.socket.ServerSocketWrapper;
import org.jboss.remoting.transport.socket.ServerThread;
import org.jboss.remoting.transport.socket.SocketServerInvoker;
import org.jboss.util.id.GUID;


/**
 * Versioning tests for JBREM-1144.
 * 
 * @author <a href="mailto:ron.sigal@jboss.com">Ron Sigal</a>
 * @version $Rev$
 * <p>
 * Copyright Nov 17, 2010
 * </p>
 */
public class ServerIdentityVersionTestServer extends ServerTestCase
{
   public static int PORT = 6543;

   private static Logger log = Logger.getLogger(ServerIdentityVersionTestServer.class);

   protected String host;
   protected int port;
   protected InvokerLocator serverLocator;
   protected Connector connector;
   protected TestInvocationHandler invocationHandler;

   public static void main(String[] args)
   {
      try
      {
         final ServerIdentityVersionTestServer p = new ServerIdentityVersionTestServer();
         Thread.sleep(300000);
         p.tearDown();
      }
      catch (Exception e)
      {
         log.error("Error", e);
      }
   }

   public void setUp() throws Exception
   {
      new Thread()
      {
         public void run()
         {
            try
            {
               Logger.getLogger("org.jboss.remoting").setLevel(XLevel.TRACE);
               Logger.getLogger("org.jboss.test.remoting").setLevel(Level.INFO);
               String pattern = "[%d{ABSOLUTE}] [%t] %5p (%F:%L) - %m%n";
               PatternLayout layout = new PatternLayout(pattern);
               ConsoleAppender consoleAppender = new ConsoleAppender(layout);
               Logger.getRootLogger().addAppender(consoleAppender);

               host = InetAddress.getLocalHost().getHostAddress();

               // Start server.
               setupServer(true);

               // Allow time to get serverId of first server.
               Thread.sleep(15000);

               // Bounce server.
               log.info("SHUTTING DOWN SERVER");
               shutdownServer();
               log.info("SHUT DOWN SERVER");
               setupServer(true);
               log.info("SET UP NEW SERVER");

               Thread.sleep(15000);
               shutdownServer();
            }
            catch (Exception e)
            {
               e.printStackTrace();
            }
         }
      }.start();
   }


   public void tearDown() throws Exception
   {
   }


   protected String getTransport()
   {
      return "socket";
   }


   protected void addExtraServerConfig(Map config) {}


   protected void setupServer(boolean useLeasing) throws Exception
   {
      String locatorURI = getTransport() + "://" + host + ":" + PORT + "/?" + Remoting.USE_SERVER_CONNECTION_IDENTITY + "=true";
      if (useLeasing)
      {
         locatorURI += "&" + InvokerLocator.CLIENT_LEASE + "=true";
         locatorURI += "&" + "leasePeriod=20000";
      }
      String metadata = System.getProperty("remoting.metadata");
      if (metadata != null)
      {
         locatorURI += "&" + metadata;
      }
      serverLocator = new InvokerLocator(locatorURI);
      log.info("Starting remoting server with locator uri of: " + locatorURI);
      HashMap config = new HashMap();
      config.put(InvokerLocator.FORCE_REMOTE, "true");
      addExtraServerConfig(config);
      connector = new Connector(serverLocator, config);
      connector.create();
      invocationHandler = new TestInvocationHandler();
      connector.addInvocationHandler("test", invocationHandler);
      if (useLeasing)
      {
         connector.addConnectionListener(new TestConnectionListener());
      }
      connector.start();
   }


   protected void shutdownServer() throws Exception
   {
      if (connector != null)
      {
         // Remoting versions 1.x and 2.2.x don't necessarily shut down all of their
         // ServerThreads, so an exiting connection could connect to a ServerThread
         // associated with the old ServerInvoker.
         SocketServerInvoker invoker = (SocketServerInvoker) connector.getServerInvoker();
         Field clientpoolField = SocketServerInvoker.class.getDeclaredField("clientpool");
         clientpoolField.setAccessible(true);
         Field socketWrapperField = ServerThread.class.getDeclaredField("socketWrapper");
         socketWrapperField.setAccessible(true);
         LRUPool clientpool = (LRUPool) clientpoolField.get(invoker);
         Set threads = clientpool.getContents();
         threads = new HashSet(threads);
         log.info("NUMBER of SERVER THREADS: " + threads.size());
         Iterator it = threads.iterator();
         while (it.hasNext())
         {
            ServerThread t = (ServerThread) it.next();
            log.info("CLOSING " + t);
            ServerSocketWrapper socketWrapper = (ServerSocketWrapper) socketWrapperField.get(t);
            socketWrapper.close();
            log.info(this + " CLOSED " + socketWrapper);
         }
         log.info("CLOSED SERVER THREADS");
         connector.stop();
         log.info("STOPPED CONNECTOR");
      }
   }


   static class TestInvocationHandler implements ServerInvocationHandler
   {
      public void addListener(InvokerCallbackHandler callbackHandler) {}
      public Object invoke(final InvocationRequest invocation) throws Throwable
      {
         return invocation.getParameter();
      }
      public void removeListener(InvokerCallbackHandler callbackHandler) {}
      public void setMBeanServer(MBeanServer server) {}
      public void setInvoker(ServerInvoker invoker) {}
   }

   static class TestConnectionListener implements ConnectionListener
   {
      public boolean connectionFailed;
      public Throwable throwable;

      public void handleConnectionException(Throwable throwable, Client client)
      {
         connectionFailed = true;
         this.throwable = throwable;
         log.info(this + " received connection notification: connectionFailed: " + connectionFailed);
      }
   }
}