/*
* JBoss, Home of Professional Open Source
* Copyright 2010, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.test.remoting.versioning.identity;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.jboss.logging.XLevel;
import org.jboss.remoting.Client;
import org.jboss.remoting.ConnectionListener;
import org.jboss.remoting.ConnectionValidator;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.Remoting;


/**
 * Versioning Unit tests for JBREM-1144.
 * 
 * @author <a href="mailto:ron.sigal@jboss.com">Ron Sigal</a>
 * @version $Rev$
 * <p>
 * Copyright Nov 17, 2010
 * </p>
 */
public class ServerIdentityWithLeasingVersionTestClient extends ServerIdentityVersionTestClientParent
{
   private static Logger log = Logger.getLogger(ServerIdentityWithLeasingVersionTestClient.class);
   private String host;
   private boolean clientImplementsServerIdentity;
   private boolean serverImplementsServerIdentity;
   protected ByteArrayOutputStream baosOut;
   protected PrintStream originalOutPrintStream;
   protected ByteArrayOutputStream baosErr;
   protected PrintStream originalErrPrintStream;
   
   public void setUp() throws Exception
   {
      clientImplementsServerIdentity = Boolean.getBoolean("clientImplementsServerIdentity");
      serverImplementsServerIdentity = Boolean.getBoolean("serverImplementsServerIdentity");
      host = InetAddress.getLocalHost().getHostAddress();

      if (clientImplementsServerIdentity && serverImplementsServerIdentity)
      {
         originalOutPrintStream = System.out;
         baosOut = new ByteArrayOutputStream();
         PrintStream ps = new PrintStream(baosOut);
         setOut(ps);
         
         originalErrPrintStream = System.err;
         baosErr = new ByteArrayOutputStream();
         ps = new PrintStream(baosErr);
         setErr(ps);
      }
      
      Logger.getLogger("org.jboss.remoting").setLevel(XLevel.TRACE);
      Logger.getLogger("org.jboss.test.remoting").setLevel(Level.INFO);
      String pattern = "[%d{ABSOLUTE}] [%t] %5p (%F:%L) - %m%n";
      PatternLayout layout = new PatternLayout(pattern);
      ConsoleAppender consoleAppender = new ConsoleAppender(layout);
      Logger.getRootLogger().addAppender(consoleAppender);

      log.info(this + " clientImplementsServerIdentity: " + clientImplementsServerIdentity);
      log.info(this + " serverImplementsServerIdentity: " + serverImplementsServerIdentity);  
      log.info(this + " host: " + host);
   }

   
   public void tearDown()
   {
   }
   
   
   public void testServerIdentityWithLeasing() throws Throwable
   {
      log.info("entering " + getName());
      
      InvokerLocator clientLocator = new InvokerLocator(createLocatorURI());
      log.info(this + "clientLocator: " + clientLocator);
      HashMap clientConfig = new HashMap();
      clientConfig.put(InvokerLocator.FORCE_REMOTE, "true");
      addExtraClientConfig(clientConfig);
      Client client = new Client(clientLocator, clientConfig);
      client.connect();
      log.info("client is connected");
      
      // Test connection.
      assertEquals("abc", client.invoke("abc"));
      log.info("connection is good");
      
      // Run appropriate test.
      if (clientImplementsServerIdentity && serverImplementsServerIdentity)
      {
         assertTrue(doServerIdentitySupportedTest(client));
      }
      else
      {
         assertTrue(doServerIdentityUnsupportedTest(client));
      }

      client.disconnect();
      log.info(getName() + " PASSES");
   }
   
   
   protected String getTransport()
   {
      return "socket";
   }
   
   
   protected void addExtraClientConfig(Map config) {}
   

   protected String createLocatorURI() throws UnknownHostException
   {
      String locatorURI = getTransport() + "://" + host + ":" + ServerIdentityVersionTestServer.PORT + "/?" + Remoting.USE_SERVER_CONNECTION_IDENTITY + "=true";
      locatorURI += "&" + InvokerLocator.CLIENT_LEASE + "=true";
      locatorURI += "&" + InvokerLocator.CLIENT_LEASE_PERIOD + "=20000";
      String metadata = System.getProperty("remoting.metadata");
      if (metadata != null)
      {
         locatorURI += "&" + metadata;
      }
      return locatorURI;
   }
   
   
   protected boolean doServerIdentitySupportedTest(Client client) throws Exception
   {
      log.info("running server identity supported test");
      
      // Install connection listener.
      TestConnectionListener listener = new TestConnectionListener();
      HashMap metadata = new HashMap();
      metadata.put(Remoting.USE_SERVER_CONNECTION_IDENTITY, "true");
      metadata.put(ConnectionValidator.VALIDATOR_PING_PERIOD, "10000");
      metadata.put(ConnectionValidator.VALIDATOR_PING_TIMEOUT, "10000");
      client.addConnectionListener(listener, metadata);
      log.info(this + " added connection listener: " + listener);
      
      // Allow time to get serverId of first server.
      Thread.sleep(12000);
      
      // Verify listener is notified if server bounces (assuming the server identity
      // facility is available.
      Thread.sleep(10000);
      log.info(this + " listener.connectionFailed: " + listener.connectionFailed);
      
      setOut(originalOutPrintStream);
      String sOut = new String(baosOut.toByteArray());
      System.out.println(sOut);
      setErr(originalErrPrintStream);
      String sErr = new String(baosErr.toByteArray());
      System.out.println(sErr);
      
      return listener.connectionFailed && (sOut.indexOf("detected new serverId:") > -1 || sErr.indexOf("detected new serverId:") > -1);

   }
   
   
   protected boolean doServerIdentityUnsupportedTest(Client client) throws Exception
   {
      log.info("running server identity unsupported test");
      
      // Install connection listener.
      TestConnectionListener listener = new TestConnectionListener();
      HashMap metadata = new HashMap();
      metadata.put(Remoting.USE_SERVER_CONNECTION_IDENTITY, "true");
      metadata.put(ConnectionValidator.VALIDATOR_PING_PERIOD, "1000");
      metadata.put(ConnectionValidator.VALIDATOR_PING_TIMEOUT, "1000");
      client.addConnectionListener(listener, metadata);
      log.info(this + " added connection listener: " + listener);
      
      // Allow ConnectionValidator to run for a while.
      Thread.sleep(10000);
      log.info(this + " listener.connectionFailed: " + listener.connectionFailed);
      return !listener.connectionFailed;
   }
   
   
   static class TestConnectionListener implements ConnectionListener
   {
      public boolean connectionFailed;
      public Throwable throwable;
      
      public void handleConnectionException(Throwable throwable, Client client)
      {
         connectionFailed = true;
         this.throwable = throwable;
         log.info(this + " received connection notification: connectionFailed: " + connectionFailed);
      }
      
   }
}